<?php

# Define a setup function
$wgExtensionFunctions[] = 'wfWgraph_Setup';
# Add a hook to initialise the magic word
$wgHooks['LanguageGetMagic'][]       = 'wfWgraph_Magic';

define('WGRAPH_VERSION', '0.8');               // current version 

$wgExtensionCredits['parserhook'][] = array(
	'name' => 'Wgraph',
	'author' 		=>  '[http://de.wikipedia.org/wiki/Benutzer:Algorithmix Algorithmix]',
	'url' 			=>  'http://www.mediawiki.org/wiki/Extension:DynamicPageList',
	'description' 	=>  'Graph generation. Produces png and svg from textual description of nodes and edges. '.
						'Can handle big, complex graphs. Can handle subgraphs and has many layout options. '.
						'For manual and live demo see the [http://semeb.com/dpldemo/Wgraph Wgraph homepage]', 
  	'version' 		=>  WGRAPH_VERSION
  );
 
function wfWgraph_Setup() {
        global $wgParser;
        # Set a function hook associating the "example" magic word with our function
        $wgParser->setFunctionHook( 'wgraph', 'wfWgraph_Render' );
}

function wfWgraph_Magic( &$magicWords, $langCode ) {
        # Add the magic word
        # The first array element is case sensitive, in this case it is not case sensitive
        # All remaining elements are synonyms for our parser function
        $magicWords['wgraph'] = array( 0, 'wgraph' );
        # unless we return true, other parser functions extensions won't get loaded.
        return true;
}

function wfWgraph_Render( &$parser ) {

		// try to guess the server operating system
		if (strpos(getenv("PATH"),'/usr/bin')===false) 	$operatingSystem = 'windows';
		else											$operatingSystem = 'linux';	
		
		global $wgServer,$wgScriptPath,$wgUploadDirectory, $wgUploadPath, $wgTitle;
		
		# The parser function itself
        # The input parameters are wikitext with templates expanded
        # The output should be wikitext too
        
        $params = array();
		$input="";
		$numargs = func_num_args();
		if ($numargs < 1) return ('!! #Wgraph: no graph description found. !!');	
		// fetch all user-provided arguments (skipping $parser)
		$arg_list = func_get_args();
		
	  	// look for options in the remaining parameters^
	  	$resolution = 150;
	  	$graphFile  = str_replace(' ','_',$wgTitle->getText());
	  	$showWGL    = false;
	  	$showGDL    = false;
	  	$svgGraph   = false;
	  	$pngGraph   = false;
	  	$thumbResolution = 0;
		for ($i = 1; $i < ($numargs-1); $i++) {
			$token = preg_split("/ *= */",trim($arg_list[$i]),2);
			if ($token[0]=='') continue;
			if (count($token)== 1) {
				if      ($token[0]=='wgl') 		$showWGL  = true;
				else if ($token[0]=='gdl')  	$showGDL  = true;
				else if ($token[0]=='png')  	$pngGraph = true;
				else if ($token[0]=='svg')  	$svgGraph = true;
				else	return("!! #Wgraph: invalid parameter: $token[0] !!");
				continue;
			}
			if      ($token[0]=='size')  $resolution=$token[1];
			else if ($token[0]=='thumb') $thumbResolution=$token[1];
			else if ($token[0]=='name')  $graphFile=$token[1];
			else	return("!! #Wgraph: invalid parameter: $token[0] !!");	
		}
		
		// the last argument is the graph definition
		
	  	$body = "graph \"$graphFile\" . {\n  ".$arg_list[--$numargs]."\n}";
	  	
		
		// adjust resolution
    	if (!preg_match("/^[1-9][0-9]+$/",$resolution)) 	$resolution = 150;
    	if (!preg_match("/^[1-9][0-9]+$/",$thumbResolution)) $thumbResolution = 0;
    	if ($thumbResolution==0)   $thumbResolution = $resolution / 10;
		if ($thumbResolution < 20) $thumbResolution = 20;
		if ($thumbResolution > $resolution) $resolution = $thumbResolution;

		// check filename
		$graphFile = str_replace(" ","_",$graphFile);
    	if (!preg_match("/^[a-zA-Z0-9][-_a-zA-Z0-9]*$/",$graphFile)) {
			return('!! #Wgraph: invalid filename, only a-zA-Z0-9 and "-,_" are allowed. !!');
		}

		// parse body
        $errorText = "";
        $aiSeeParser = new AiSeeParser();
		$gdlText = $aiSeeParser->parse($graphFile, $body, $errorText);
		
		// write GDL output to file in /images directory
        $workDir = "$wgUploadDirectory/Wgraph";
        $binDir  = "$workDir/../../extensions/Wgraph";
        $imgPath = "$wgUploadPath/Wgraph";
        $gdlFile = fopen("$workDir/$graphFile.gdl","w");
        fputs($gdlFile,$gdlText);
        fclose($gdlFile);
        
		global $wgRawHtml;
		$wgRawHtml = true;
        $graph='';
        if ($errorText=='') {
	        $graph = '<html>';
	        if ($pngGraph || $svgGraph) {
		        // currently we produce svg and png together which is rather inefficient
		        
		        if ($operatingSystem=='windows') {
			        $wgraph = str_replace('/','\\',"$binDir/windows/makegraph.bat");
        		} else { 
			        $wgraph = "$binDir/linux/makegraph.sh";
		    	}
    		    // call a batch file or a shell script to recode (u8..iso_8859-1) & generate the graph using aiSee
    	   	    system ("$wgraph $graphFile $thumbResolution $resolution");
    	    }
    	    $noCache = "?" . time(); // we add a marker to avoid browser image caching
	        if ($pngGraph) {
				$graph .= '<a href="'."$imgPath/$graphFile".'.png' . $noCache . '" target=wgraph>'.
        				 '<img src="'."$imgPath/$graphFile" . '_thumb.png' . $noCache . '" title="click to enlarge .."></a>';
			}
	        if ($svgGraph && !$pngGraph ) {
		        $graph .= '<a href="'."$imgPath/$graphFile".'.svg' . $noCache . '" target=wgraph>'.
        				 	 '<img src="'."$imgPath/$graphFile" . '_thumb.png' . $noCache . '" title="click for SVG .."></a>';
			}
	        if (!$pngGraph && !$svgGraph) {
				$graph .= '<a href="'."$imgPath/$graphFile".'.gdl">'.$graphFile.'.gdl</a>';
	        }
			$graph .= '<small><p style="text-align:right;">';
       		if ($pngGraph && $svgGraph) {
		        $graph .= '<a href="'."$imgPath/$graphFile".'.svg' . $noCache . '" target=wgraph><i>svg</i></a> &nbsp; ';
	       	}
	       	$graph .= '<i>(made with <a href=http://semeb.com/dpldemo/Wgraph target=wgraph>Wgraph</a>)</i>';
    		$graph .= "</p></small>\n";
	        $graph .= '</html>';
    	}

    	$result=$graph;

    	if ($errorText!='') {
			$result .= "<pre>$errorText</pre>\n";
			$showWGL = true;
		}
		
        if ($showWGL) {
		    $bodyLines = split("\n",$body);
    		foreach ($bodyLines as $nr => $line) {
	    		$bodyLines[$nr] = str_pad(($nr+1),3," ",STR_PAD_LEFT).": $bodyLines[$nr]";
    		}
    		$result .= "<pre>".join ("\n",$bodyLines)."</pre>\n";
		}

        if ($showGDL) {
			$result.= "\n----\n<pre>$gdlText</pre>";
		}
		
		return $result;
}


class AiSeeParser {
	
	private $enums = array ( 	"color" 		=> "*",
								"foldstate" 	=> "*",
								"shape" 		=> "*",
								"arrowstyle" 	=> "*",
								"arrowmode" 	=> "*",
								"edgekind"	 	=> "*",
								"linestyle"		=> "*",
								"layout" 		=> "*",
								"orientation" 	=> "*",
								"magfield"	 	=> "*",
								"valign"	 	=> "*",
								"halign"	 	=> "*",
								"viewtype"	 	=> "*",
								"cr_weight" 	=> "*",
							);
							
	private $enum = array (
		'boolean' 	 => array (	'yes' => 'yes', 'true' => 'yes', '1' => 'yes', 
								'no' => 'no', 'false' => 'no', '0' => 'no' 
							  ),
		'color' 	 => array ( 'white' 		=> '*',
								'blue' 			=> '*',
								'red' 			=> '*',
								'green' 		=> '*',
								'yellow' 		=> '*',
								'magenta' 		=> '*',
								'cyan' 			=> '*',
								'darkgray' 		=> '*',
								'darkgrey' 		=> '*',
								'darkblue' 		=> '*',
								'darkred' 		=> '*',
								'darkgreen' 	=> '*',
								'darkyellow' 	=> '*',
								'darkmagenta'	=> '*',
								'darkcyan'		=> '*',
								'gold'	 		=> '*',
								'lightgray'		=> '*',
								'lightgrey'		=> '*',
								'lightblue'		=> '*',
								'lightred' 		=> '*',
								'lightgreen'	=> '*',
								'lightyellow'	=> '*',
								'lightmagenta'	=> '*',
								'lightcyan'		=> '*',
								'lilac' 		=> '*',
								'turquoise'		=> '*',
								'aquamarine'	=> '*',
								'khaki'			=> '*',
								'purple'		=> '*',
								'yellowgreen'	=> '*',
								'pink'	 		=> '*',
								'orange'		=> '*',
								'orchid' 		=> '*',
								'black' 		=> '*',
							  ),
		'edgekind'	=> array ( 	'normal'		=> '*',
								'back'			=> '*',
								'near' 			=> '*',
								'rightnear'		=> '*',
								'leftnear'		=> '*',
								'bentnear' 		=> '*',
								'rightbentnear'	=> '*',
								'leftbentnear'	=> '*',
							  ),
		'foldstate'	=> array (	'unfolded'		=> '*',
								'folded' 		=> '*',
								'boxed' 		=> '*',
								'clustered'		=> '*',
								'wrapped' 		=> '*',
								'exclusive' 		=> '*',
							  ),
		'shape' 	 => array (	'box' 			=> '*',
								'hexagon' 		=> '*',
								'lparallelogram'=> '*',
								'rparallelogram'=> '*',
								'trapezoid'		=> '*',
								'uptrapezoid'	=> '*',
								'rhomboid' 		=> '*',
								'rhomb' 		=> '*',
								'ellipse' 		=> '*',
								'circle' 		=> '*',
								'triangle' 		=> '*',
							  ),
		'arrowstyle' => array (	'solid' 		=> '*',
								'line'			=> '*',
								'none'			=> '*',
							  ),
		'arrowmode' => array  (	'free'	 		=> '*',
								'fixed'			=> '*',
							  ),
		'linestyle'	=> array  (	'continuous'	=> '*',
								'dashed'		=> '*',
								'dotted'		=> '*',
								'double'		=> '*',
								'invisible'		=> '*',
								'solid'			=> '*',
								'triple'		=> '*',
							  ),
		'layout'	 => array (	'normal' 		=> '*',
								'maxdepth'		=> '*',
								'mindepth'		=> '*',
								'dfs'			=> '*',
								'maxdepthslow'	=> '*',
								'mindepthslow'	=> '*',
								'maxindegree'	=> '*',
								'minindegree'	=> '*',
								'maxoutdegree'	=> '*',
								'minoutdegree'	=> '*',
								'maxdegree'		=> '*',
								'mindegree'		=> '*',
								'minbackward'	=> '*',
								'tree'			=> '*',
								'forcedir'		=> '*',
								'maxdepth'		=> '*',
								'maxdepth'		=> '*',
								'maxdepth'		=> '*',
							  ),
		'orientation'=> array (	'top_to_bottom'	=> '*',
								'bottom_to_top'	=> '*',
								'left_to_right'	=> '*',
								'right_to_left'	=> '*',
							  ),
		'magfield'	=> array  (	'no'			=> '*',
								'top_to_bottom'	=> '*',
								'bottom_to_top'	=> '*',
								'left_to_right'	=> '*',
								'right_to_left'	=> '*',
								'polar'			=> '*',
								'circular'		=> '*',
								'orthogonal'	=> '*',
								'polcircular'	=> '*',
							  ),
		'valign'	=> array  (	'top'			=> '*',
								'center'		=> '*',
								'bottom'		=> '*',
							  ),
		'halign'	=> array  (	'center'		=> '*',
								'left'			=> '*',
								'left_justify'	=> '*',
								'right'			=> '*',
								'right_justify'	=> '*',
							  ),
		'viewtype'	=> array  (	'normal'		=> '*',
								'cfish'			=> '*',
								'fcfish'		=> '*',
								'dcfish'		=> '*',
								'pfish'			=> '*',
								'dpfish'		=> '*',
							  ),
		'cr_weight'	 => array (	'bary'			=> '*',
								'median'		=> '*',
								'barymedian'	=> '*',
								'medianbary'	=> '*',
							  ),
	);
		 
	private $cmdWordsWithIndex = array (
		'infoname'			=> true,
		'useractionname'	=> true,
		'colorentry'		=> true,
		'classname'			=> true,
		
	);

	private $cmdWordsWithBody = array (
		'graph'				=> true,  // id  [graph]
		'subgraph'			=> true,  // id  [graph]
		'node'				=> true,  // id  [graph]
		'region'			=> true,  // id
		'nodetype'			=> true,  // id
		'edgetype'			=> true,  // id
		'edge'				=> true,  // (from) source (to) target
		'backedge'			=> true,
		'nearedge'			=> true,
		'rightnearedge'		=> true,
		'leftnearedge'		=> true,
		'bentnearedge'		=> true,
		'rightbentnearedge'	=> true,
		'leftbentnearedge'	=> true,
	);

	private $cmdWordHasArgumentType = array (

//graph
		'align'					=> 'halign',		// alias for textmode
		'amax'					=> 'unsigned',
		'arrow_mode'			=> 'arrowmode',
		'attraction'			=> 'unsigned',
		
		'bmax'					=> 'unsigned',
		'border'				=> 'xy_pair',
		'bordercolor'			=> 'color',			// graph, node, nodetype
		'borderstyle'			=> 'linestyle',		// subgraph, node, nodetype
		'borderwidth'			=> 'unsigned',		// subgraph, node, nodetype
		
		'classname'				=> 'string',		// index
		'cmin'					=> 'unsigned',
		'cmax'					=> 'unsigned',
		'color'					=> 'color',			// graph, subgraph, node, edge, nodetype, edgetype
		'colorentry'			=> 'colorset',		// index
		'crossing_optimization'	=> 'boolean',
		'crossing_phase2'		=> 'boolean',
		'crossing_weight'		=> 'cr_weight',
		
		'dirty_edge_labels'		=> 'boolean',
		'display_edge_labels'	=> 'boolean',
		
		'edges'					=> 'boolean',
		'energetic'				=> 'boolean',
		'energetic_attraction'	=> 'float',
		'energetic_repulsion'	=> 'float',
		'energetic_gravity'		=> 'float',
		'energetic_crossing'	=> 'float',
		'energetic_overlapping'	=> 'float',
		'energetic_border'		=> 'float',
		'equal_y_dist'			=> 'boolean',
		
		'fast_icons'			=> 'boolean',
		'fdmax'					=> 'unsigned',
		'finetuning'			=> 'boolean',
		'focus'					=> 'none',			// graph, node
		'font'					=> 'string',		// alias for fontname
		'fontname'				=> 'string',		// graph, node, nodetype
		'fstraight_phase'		=> 'boolean',
		
		'gravity'				=> 'float',
		
		'height'				=> 'unsigned',		// graph, node, nodetype
		'hidden'				=> 'unsigned',
		'horizontal_order'		=> 'unsigned',		// graph, node, edge, nodetype, edgetype
		
		'iconcolors'			=> 'unsigned',
		'icon'					=> 'string',		// alias for iconfile
		'iconfile'				=> 'string',		// graph, node, nodetype
		'icons'					=> 'boolean',
		'ignore_singles'		=> 'boolean',
		'importance'			=> 'unsigned',		// graph, node, nodetype
		'info1'					=> 'string',		// graph, node, nodetype
		'info2'					=> 'string',		// graph, node, nodetype
		'info3'					=> 'string',		// graph, node, nodetype
		'infoname'				=> 'string',		// index
		'inport_sharing'		=> 'boolean',
		'invisible'				=> 'boolean',
		
		'label'					=> 'string',		// graph, node, edge, nodetype, edgetype
		'late_edge_labels'		=> 'boolean',
		'layoutalgorithm'		=> 'layout',
		'layout_downfactor'		=> 'unsigned',
		'layout_upfactor'		=> 'unsigned',
		'layout_nearfactor'		=> 'unsigned',
		'level'					=> 'unsigned',
		'linear_segments'		=> 'boolean',
		'loc'					=> 'xy_pair',		// graph, node
		
		'magnetic_field1'		=> 'magfield',
		'magnetic_field2'		=> 'magfield',
		'magnetic_force1'		=> 'unsigned',
		'magnetic_force2'		=> 'unsigned',
		'manhattan_edges'		=> 'boolean',
		
		'near_edges'			=> 'boolean',
		'nodes'					=> 'boolean',
		'node_alignment'		=> 'valign',
		
		'orientation'			=> 'orientation',
		'outport_sharing'		=> 'boolean',
		
		'pmin'					=> 'unsigned',
		'pmax'					=> 'unsigned',
		'portsharing'			=> 'boolean',
		'priorityphase'			=> 'boolean',
		
		'randomfactor'			=> 'unsigned',
		'randomimpulse'			=> 'unsigned',
		'randomrounds'			=> 'unsigned',
		'repulsion'				=> 'unsigned',
		'rmin'					=> 'unsigned',
		'rmax'					=> 'unsigned',
		
		'scaling'				=> 'float',			// graph, node, nodetype
		'shape'					=> 'shape',			// graph, node, nodetype
		'shrink'				=> 'unsigned',		// graph, node, nodetype
		'smanhattan_edges'		=> 'boolean',
		'smax'					=> 'unsigned',
		'splinefactor'			=> 'unsigned',
		'splines'				=> 'boolean',
		'spreadlevel'			=> 'unsigned',
		'state'					=> 'foldstate',     // graph, node, region
		'stretch'				=> 'unsigned',		// graph, node, nodetype
		'straightphase'			=> 'boolean',
		'subgraph_labels'		=> 'boolean',
		
		'tempfactor'			=> 'float',
		'tempmin'				=> 'unsigned',
		'tempmax'				=> 'unsigned',
		'tempscheme'			=> 'unsigned',
		'temptreshold'			=> 'unsigned',
		'textcolor'				=> 'color',			// graph, node, edge, nodetype, edgetype
		'textmode'				=> 'halign',		// graph, node, nodetype
		'title'					=> 'string',		// graph, node
		'treefactor'			=> 'float',
		'useractioncmd1'		=> 'string',
		'useractioncmd2'		=> 'string',
		'useractioncmd3'		=> 'string',
		'useractioncmd4'		=> 'string',
		'useractionname'		=> 'string',   // index
		
		'vertical_order'		=> 'unsigned',		// graph, node, nodetype
		'view'					=> 'viewtype',
		
		'width'					=> 'unsigned',		// graph, node, nodetype
		
		'x'						=> 'unsigned',
		'xbase'					=> 'unsigned',
		'xmax'					=> 'unsigned',
		'xlspace'				=> 'unsigned',
		'xspace'				=> 'unsigned',
		
		'y'						=> 'unsigned',
		'ybase'					=> 'unsigned',
		'ymax'					=> 'unsigned',
		'yspace'				=> 'unsigned',
		
// region
		'sources'				=> 'objlist',
		'targets'				=> 'objlist',
		'range'					=> 'unsigned',
		'use_edgetypes'			=> 'numberlist',	// enumeration of edgetypes to use

// edge
		'anchor'				=> 'unsigned',
		'arrowcolor'			=> 'color',
		'arrowsize'				=> 'unsigned',
		'arrowstyle'			=> 'arrowstyle',
		'backarrowcolor'		=> 'color',
		'backarrowsize'			=> 'unsigned',
		'backarrowstyle'		=> 'arrowstyle',
		'class'					=> 'unsigned',
		'linestyle'				=> 'linestyle',
		'priority'				=> 'unsigned',
		'sourcename'			=> 'string',	// alias for source
		'source'				=> 'string',
		'from'					=> 'string',	// alias for source
		'targetname'			=> 'string',	// alias for target
		'target'				=> 'string',
		'to'					=> 'string',	// alias for target
		'thickness'				=> 'unsigned',
		
// additional (not aiSee)
		'edgetypes'				=> 'typelist',	// graph: priority sequence
		'nodetypes'				=> 'typelist',	// graph        comma separated priority sequence
		'show_nodes'			=> 'objlist',   // graph  only these are displayed
		'skip_nodes'			=> 'objlist',   // graph  thrown away
		'hide_nodes'			=> 'objlist',   // graph  invisible, still influencing layout
		'show_nodetypes'		=> 'typelist',  // graph  only these are displayed
		'show_edgetypes'		=> 'typelist',  // graph
		'skip_nodetypes'		=> 'typelist',  // graph  thrown away
		'skip_edgetypes'		=> 'typelist',  // graph
		'hide_nodetypes'		=> 'typelist',  // graph  invisible, still influencing layout
		'hide_edgetypes'		=> 'typelist',  // graph
		'kind'					=> 'edgekind',	// edge
		'type'					=> 'typelist',	// node, edge   comma separated list of types
		'partof'				=> 'string',	// node & graph define hierarchy
		'layer'					=> 'string',	// node
		'url'					=> 'string',	// node  , becomes URL-escaped
		'href'					=> 'string',	// alias for 'url'
		'textwidth'				=> 'unsigned',	// node, edge
		'weight'				=> 'unsigned',	// nodetype, edgetype  defines a default rank if edgetypes/modetypes were not specified
				
	);
		
	private $state;
	private $level;
	private $edgeKind;
	
	private $text;
	private $textLen;
	private $pos,$errorPos;
	private $line,$lineStartPos;

	private $outText;
	private $errorFound = false;
	public  $errorText;	
	
	private $colorMap = array();
	private $colorMapIndex=31;
	
	function loadText($text) {
		$this->text			= $text;
		$this->pos			= -1;
		$this->errorPos		= "";
		$this->line			= 0;
		$this->lineStartPos	= 0;
		$this->textLen		= strlen($text);
	}
	
	function endOfText() {
		return ($this->pos >= $this->textLen);
	}

	function getChar() {
		if ( $this->pos + 1 >= $this->textLen) return -1;
		++$this->pos;
		$c = $this->text[$this->pos];
		if ($c=="\n") {
			++$this->line;
			$this->lineStartPos = $this->pos;
		}
		return $c;
	}

	function ungetChar($n=1) {
		for ($i=0;$i<$n;$i++) {
			if (($this->pos) < $this->textLen) {
				if ($this->text[$this->pos] == "\n") --$this->line;
				--$this->pos;
			}
		}
	}

	function eatSpace() {
		// remove space and comma, /* .. */ and // ....\n
		$comment = 0;
		while (($c=$this->getChar()) != -1) {
			if ($comment==0) {
				if (ctype_space($c) || $c==',') {
					// do nothing
				} else if ($c=='/') {
					$comment = 1;
				} else {
					$this->ungetChar();
					return true;
				}
			} else if ($comment==1) {
				if 		  ($c=='/') {
					$comment = 2;
				} else if ($c=='*') {
					$comment = 3;
				} else {
					$this->ungetChar(2);
					return true;
				}
			} else if ($comment==2) {
				if ($c=="\n") {
					$comment = 0;
				}
				else {
				}
			} else if ($comment==3) {
				if ($c=='*') {
					$comment = 4;
				}
				else {
				}
			} else if ($comment==4) {
				if ($c=='/') $comment = 0;
				else		 $comment = 3;
			}
		}
		return false;
	}

	function eatToken($msg) {
		$this->setError();
		$token="";
		while (($c=$this->getChar()) != -1) {
			if (!ctype_space($c) && $c!= ',') {
				$token .= $c;
			}
			else {
				$this->ungetChar();
				$this->error($msg,$token);
				return;
			}
		}
		$this->error($msg,$token);
	}

	// we accept "=" and ":"
	function eatColon($obligatory=true, $waitForBlock=false) {
		if (($c=$this->getChar()) != -1) {
			if ($c==':'||$c=='=') return;
			$this->ungetChar();
			if ($c=='{') return;
			if (!$obligatory) return;
			if ($waitForBlock)	$this->eatToken("expecting '{'");
			else 				$this->eatToken("expecting ':' or '='.");
		}
	}

	function eatBraceOpen() {
		if (($c=$this->getChar()) != -1) {
			if ($c=='{') {
				// print "BLOCK {\n";
				return;
			}
			$this->ungetChar();
			$this->eatToken("expecting '{'.");
		}
	}

	function getCmdWord() {
		if (($cmdWord=$this->getChar()) != -1) {
			if (ctype_alpha($cmdWord)) {
				while (($c=$this->getChar()) != -1) {
					if (ctype_alnum($c) || $c=='_' || $c==".") {
						$cmdWord .= $c;
					}
					else {
						$this->ungetChar();
						// print "CMD=$cmdWord\n";
						return strtolower($cmdWord);
					}
				}
			}
			else if ( $cmdWord=='}' ) {
				// print "BLOCK }\n";
				return $cmdWord;
			}	
			else {
				$this->ungetChar();
				$this->eatToken("command word expected");
				return "";
			}
		}
	 }

	function getArgument($type) {

		$argument="";
		
		if ($type=='token') {
			while (($c=$this->getChar()) != -1) {
				if (ctype_alnum($c) || $c=='_' || $c=='#') {
					$argument .= $c;
				}
				else {
					$this->ungetChar();
					return $argument;
				}
			}

		} else if ($type=='string') {
			if (($c=$this->getChar()) != -1) {
				if ($c=='"' || $c == "'") {
					// delimiters used
					$delim = $c;
					$lastC='';
					while (($c=$this->getChar()) != -1) {
						if ($c != $delim || $lastC=='\\') {
							if ($delim =="'" && $c=='"') $argument .= '\\';
							$argument .= $c;
							$lastC = $c;
						}
						else {
							// look for more strings and concatenate
							$pos = $this->pos;
							$this->eatSpace();
							if (($c=$this->getChar()) == $delim) {
								$this->ungetChar();
								$argument = trim($argument) . $this->getArgument("string");
							}
							else {
								$this->ungetChar($this->pos-$pos);
							}
							return str_replace('_',' ',trim($argument));
						}
					}
					$this->setError();
					$this->error("undelimited string",$argument);
				}
				else {
					// no delimiters used
					$argument=$c;
					while (($c=$this->getChar()) != -1) {
						if (ctype_space($c) || $c == ',' || $c == '{' || $c == '}') {
							// break at space, comma  or brace
							$this->ungetChar();
							return $argument;
						}
						if ($c == '_') $c=' ';
						$argument .= $c;
					}
					return $argument;
				}
			}

		} else if ($type=="boolean") {
			$argument = $this->getArgument("token");
			if (array_key_exists(strtolower($argument),$this->enum['boolean']) ) {
				return $this->enum["boolean"][strtolower($argument)];
			}
			else {
				$this->setError();
				$this->error("unknown value, allowed values are:".join(',',array_keys($this->enum["boolean"])),$argument);
				return "";
			}
			
		} else if ($type=="typelist" || $type=="objlist" || $type=="numberlist") {
			return ','.preg_replace('/\\s*[,;]\\s*/',',',trim($this->getArgument("string"))) . ',';
			
		} else if (array_key_exists($type, $this->enums)) {
			$argument = $this->getArgument("token");
			if (array_key_exists(strtolower($argument),$this->enum[$type]) ) {
				return strtolower($argument);
			}
			else if ($type=='color') {
				if (preg_match('/^#[0-9a-fA-F]{6}$/',$argument)) {
					$argument = strtolower($argument);
					if (array_key_exists($argument,$this->colorMap)) return $this->colorMap[$argument];
					if ($this->colorMapIndex>=99) return 'lilac';
					$this->colorMap[$argument] = ++$this->colorMapIndex;
					return $this->colorMapIndex;
				}
				else {
					$this->setError();
					$this->error("color must be of type #000000 .. #ffffff or:\n      ".join(',',array_keys($this->enum[$type])),$argument);
					return "";
				}
			}
			else {
				$this->setError();
				$this->error("unknown value, allowed values are:".join(',',array_keys($this->enum[$type])),$argument);
				return "";
			}
			
		} else if ($type=='float') {
			while (($c=$this->getChar()) != -1) {
				if (ctype_digit($c) || $c=='.') {
					$argument .= $c;
				}
				else {
					$this->ungetChar();
					if (preg_match('/^[0-9]*[.]?[0-9]*$/',$argument) && $argument != "") {
						if (preg_match('/^[0-9]+$/',$argument)) {
							$argument .= ".0";
						}
						// print "ARG(float)=$argument\n";
						return $argument;
					}
					else {
						$this->setError();
						$this->error("invalid floating number",$argument);
						return;
					}
				}
			}
			
		} else if ($type=='unsigned') {
			while (($c=$this->getChar()) != -1) {
				if (ctype_digit($c)) {
					$argument .= $c;
				}
				else {
					$this->ungetChar();
					// print "ARG(unsigned)=$argument\n";
					return $argument;
				}
			}
			
		} else if ($type=='colorset') {
			$n = 0;
			$withinNumber = false;
			while (($c=$this->getChar()) != -1) {
				if (ctype_digit($c)) {
					$argument .= $c;
					if (!$withinNumber) { $n++; $withinNumber=true; } 
				} else if (ctype_space($c) && $n<=2) {
					$withinNumber=false;
					$argument .= " ";
				} else {
					$this->ungetChar();
					$num = preg_split('/\\s+/',$argument);
					if (count($num)==3 && $num[0]>=0 && $num[0]<=255 && $num[1]>=0 && $num[1]<=255 && $num[2]>=0 && $num[2]<=255) {
						// print "ARG(colorset)=$argument\n";
						return $argument;
					}
					else {
						$this->setError();
						$this->error("invalid color set, form must be: nnn nnn nnn",$argument);
						return;
					}
				}
			}
		}
		return "";

	}

	function getIndexNumber() {
		
		if (($index=$this->getChar()) != -1) {
			if (ctype_digit($index) && $index != '0') {
				while (($c=$this->getChar()) != -1) {
					if (ctype_digit($c)) {
						$index .= $c;
					}
					else {
						$this->ungetChar();
						// print "INDEX=$index\n";
						return $index;
					}
				}
			}
			else {
				$this->ungetChar();
				$this->eatToken("positive index number expected");
				return "";
			}
		}
	}
	
	
	//================================================================================================
	
	public function parse($graphName, $text, &$errorText) {
		$this->loadText(" $text");
		$this->state = 'EXPECT_CMDWORD';
		$gdl = new GdlTree();
		$gdl->init($this);
		$this->level=0;
		$index="";
		$partOf = $graphName;
				
		while (!$this->endOfText() && !$this->isError()) {

			if       ($this->state=='EXPECT_CMDWORD') {
				if (!$this->eatSpace()) break;
				$cmdPos=$this->getPos();
				$cmdWord = $this->getCmdWord();
				$id='';
				
				if ($cmdWord=="}") {				// close body
					if (--$this->level<0) {
						$this->setError();
						$this->error("block nesting level error",$cmdWord);
					}
					$this->resetErrorPos(); // in case the closing brace is missing
					$gdl->add(new Command("}","","","",$cmdPos));
				}	
				else {
					if ($cmdWord=="from") 		$cmdWord="source";
					if ($cmdWord=="sourcename") $cmdWord="source";
					if ($cmdWord=="to")   		$cmdWord="target";
					if ($cmdWord=="targetname")	$cmdWord="target";
					if ($this->hasBody($cmdWord)) {
						if 		($cmdWord=='edge') 				$this->edgeKind = 'normal';
						else if ($cmdWord=='backedge') 			$this->edgeKind = 'back';
						else if ($cmdWord=='nearedge')			$this->edgeKind = 'near';
						else if ($cmdWord=='rightnearedge')		$this->edgeKind = 'rightnear';
						else if ($cmdWord=='leftnearedge')		$this->edgeKind = 'leftnear';
						else if ($cmdWord=='bentnearedge')		$this->edgeKind = 'bent';
						else if ($cmdWord=='leftbentnearedge')	$this->edgeKind = 'leftbentnear';
						else if ($cmdWord=='rightbentnearedge')	$this->edgeKind = 'rightbentnear';
						if ($cmdWord == 'region') 	$this->state='START_BODY';
						else						$this->state='EXPECT_ID';
					}
					else {
						// we fetch the arg type and produce an error if the command is unknown
						$argumentType = $this->hasArgumentType($cmdWord);
						if ($this->hasIndex($cmdWord)) 	$this->state = 'EXPECT_INDEX';
						else							$this->state = 'EXPECT_ARGUMENT';
						if ($argumentType=="none") {
							$gdl->add (new Command($cmdWord,"",$argumentType,"",$cmdPos));
							$this->state = 'EXPECT_CMDWORD';
						}
					}
				}
				
			} else if ($this->state=='EXPECT_ID') {
				if (!$this->eatSpace()) break;
				if (strpos(' ,'
					.'edge,backedge,nearedge,bentnearedge,leftnearedge,rightnearedge,leftbentnearedge,rightbentnearedge,'
					,",$cmdWord,") > 0) {
					// we found an edge
					$from='';
					$to='';
					$ch= $this->getChar(); $this->ungetChar();
					if ($ch!='{') {
						// from and to given
						$from = $this->getArgument("string");
						if (strcasecmp($from,'from')==0) {
							if (!$this->eatSpace()) break;
							$from = $this->getArgument("string");
						}
						if (!$this->eatSpace()) break;
						$to = $this->getArgument("string");
						if (strcasecmp($to,'to')==0) {
							if (!$this->eatSpace()) break;
							$to = $this->getArgument("string");
						}
					}
				}
				else if (strpos(' ,node,graph,subgraph,',",$cmdWord,") > 0) {
					// graph, subgraph  or node
					$id = $this->getArgument("string");
					// $this->errorText .= "id='$id'\n";
					if (!$this->eatSpace()) break;
					// after the id we might find the id of a (sub)graph to which the current node or graph shall belong
					$partOf = $graphName;
					$ch= $this->getChar(); $this->ungetChar();
					if ($ch!='{') $partOf = $this->getArgument("string");
				}
				else {
					// nodetype or edgetype
					$id = $this->getArgument("string");
				}
				$this->state = 'START_BODY';
			
			} else if ($this->state=='EXPECT_INDEX') {
				if (!$this->eatSpace()) break;
				$index = $this->getIndexNumber();
				$this->state = 'EXPECT_ARGUMENT';
			
			} else if ($this->state=='EXPECT_ARGUMENT') {
				if (!$this->eatSpace()) break;
				$this->eatColon(false);
				if (!$this->eatSpace()) break;
				$argument = $this->getArgument($argumentType);
				$gdl->add(new Command($cmdWord,$index,$argumentType,$argument,$cmdPos));
				$this->state = 'EXPECT_CMDWORD';
				$index="";
				
			} else if ($this->state=='START_BODY') {
				if (!$this->eatSpace()) break;
				$this->eatColon(true,true);
				if (!$this->eatSpace()) break;
				$this->eatBraceOpen();
				$this->setErrorPos(); // in case the closing brace is missing
				
				// add the object to the command list
				$gdl->add(new Command($cmdWord,"","","{",$cmdPos));
				
				if (strpos(' ,node,graph,subgraph,',",$cmdWord,") > 0) {
					$gdl->add(new Command('title',"","string",$id,$cmdPos));
					$gdl->add(new Command('partof',"","string",$partOf,$cmdPos));
				} else if (strpos(' ,nodetype,edgetype,',",$cmdWord,") > 0) {
					$gdl->add(new Command('title',"","string",$id,$cmdPos));
				} else if (strpos(' ,'
					.'edge,backedge,nearedge,bentnearedge,leftnearedge,rightnearedge,leftbentnearedge,rightbentnearedge,'
					,",$cmdWord,") > 0) {
					$gdl->add(new Command('source',"","string",$from,$cmdPos));
					$gdl->add(new Command('target',"","string",$to,$cmdPos));
					$gdl->add(new Command('kind',"","edgekind",$this->edgeKind,$cmdPos));
				}
				$this->state = 'EXPECT_CMDWORD';
				++$this->level;
			}
		}

		$gdl->useColorMap($this->colorMap);	
		$gdl->add(new Command('%EOF','','intern','%EOF',''));
		
		if ($this->state != 'EXPECT_CMDWORD') {
			$this->setError();
			$this->error("incomplete command",$cmdWord);
		}

		// undefined nodes are automatically assigned to nodetype 'undefined'
		$gdl->findUndefinedNodes($graphName);

		// resolve types, apply filters etc.
		$gdl->resolveTypes($graphName);
			
		$errorText = $this->errorText;
		return $gdl->printTreeAsGdl();
	}
	
	//================================================================================================
	
	function hasIndex($cmdWord) {
		return array_key_exists($cmdWord,$this->cmdWordsWithIndex);
	}

	function hasBody($cmdWord) {
		return array_key_exists($cmdWord,$this->cmdWordsWithBody);
	}

	function hasArgumentType($cmdWord) {
		$baseCmdWord = preg_replace("/^(node|edge|foldnode|foldedge)[.]/","",$cmdWord);
		if (!array_key_exists($baseCmdWord,$this->cmdWordHasArgumentType)) {
			$this->setError();
			$this->error("unknown command",$cmdWord);
			return "";
		}
		return $this->cmdWordHasArgumentType[$baseCmdWord];
	}

	function setError($takeCurrentPosition = true)  {
		$this->errorFound=true;
		if ($takeCurrentPosition || $this->errorPos=='') $this->setErrorPos();
	}

	function setErrorPos()  {
		if ($this->errorPos == '') $this->errorPos = $this->getPos();
		else $this->errorPos .= " -- ".$this->getPos();
	}

	function resetErrorPos()  {
		if (!$this->errorFound) $this->errorPos  = '';
	}

	function getPos()  {
		return "line ".($this->line+1).", char ".($this->pos - $this->lineStartPos + 1);
	}

	function isError()  {
		return $this->errorFound;
	}
	
	function error($msg,$token) {
		$this->errorText .= "error ($this->errorPos, $this->state): $msg\n";
		$this->errorText .= "      '$token'\n";
	}
	 
}

//*******************************************************************************************************//

class GdlTree {
	private $cmds = array();

	private $parser;
	private $state;
	
	private $nodeTypes = array();
	private $edgeTypes = array();
	private $nodes     = array();         // all nodes
	private $edges     = array();         // all edges
	private $regions   = array();         // all regions
	
	private $currentNode, $currentEdge, $currentNodeType, $currentEdgeType, $currentRegion;
	private $nodeTypeSeq = array();   // preference for applying type based attributes
	private $edgeTypeSeq = array();
	private $filters = array();       // show, skip & hide nodes and edges based on their types
	private $edgeHasNode = array();  // store the nodes which are touched by an edge

	private $nodeEdgeCount = array();

	private $nodeTypeNumber = -1;
	private $edgeTypeNumber = -1;
	private $edgeClassNumber = 0;
	private $edgeNumber = -1;	
	private $regionNumber = -1;	
	
	private $edgeLabelsCmd = false;
	private $colorEntries = array();	
	
	private $partOf = array();  // subgraph nesting
	
	// initialize the syntax tree
	public function init($parser) {
		$this->state = 'BEGIN';
		$this->parser = $parser;
		for ($f=0;$f<9;$f++) $this->filters[$f]='';
	}

	// check if command is allowed in this state
	public function isValidStateForCommand($cmd,$state) {

		if ($cmd=='}' && 
				($state=='IN_GRAPH' || $state=='IN_SUBGRAPH'  || $state=='IN_REGION' || 
				 $state == 'IN_NODE' || $state=='IN_EDGE'    ||
				 $state=='IN_NODETYPE' || $state =='IN_EDGETYPE')) {
			return true;
		}
		
		if ($state == 'IN_GRAPH') {
			if (strpos(' ,'
				.'subgraph,nodetype,edgetype,'
				.'edgetypes,nodetypes,show_nodes,skip_nodes,hide_nodes,'
				.'show_nodetypes,show_edgetypes,skip_nodetypes,skip_edgetypes,hide_nodetypes,hide_edgetypes,'
				,",$cmd,")>0) return true;
			if (strpos("  $cmd"," node.")>0 ) {
				$state = 'IN_NODE';
				$cmd = substr($cmd,5);
			}
			if (strpos("  $cmd"," edge.")>0 ) {
				$state = 'IN_EDGE';
				$cmd = substr($cmd,5);
			}
		}

		// in theory IN_SUBGRAPH should also be possible here ..
		if ( ($state == 'IN_GRAPH' ) && strpos(' ,'
			.'region,node,partof,'
			.'edge,backedge,nearedge,bentnearedge,rightbentnearedge,leftbentnearedge,leftnearedge,rightnearedge,'
			.'align,amax,arrow_mode,attraction,'
			.'bmax,border,bordercolor,borderstyle,borderwidth,'
			.'classname,cmin,cmax,color,colorentry,crossing_optimization,crossing_phase2,crossing_weight,'
			.'dirty_edge_labels,display_edge_labels,'
			.'edges,energetic,energetic_attraction,energetic_repulsion,'
			.'energetic_gravity,energetic_crossing,energetic_overlapping,energetic_border,equal_y_dist,'
			.'fast_icons,fdmax,finetuning,focus,font,fontname,fstraight_phase,'
			.'gravity,'
			.'height,hidden,horizontal_order,'
			.'iconcolors,icon,iconfile,icons,ignore_singles,importance,info1,info2,info3,infoname,inport_sharing,invisible,'
			.'label,late_edge_labels,layoutalgorithm,layout_downfactor,layout_upfactor,layout_nearfactor,level,linear_segments,loc,'
			.'magnetic_field1,magnetic_field2,magnetic_force1,magnetic_force2,manhattan_edges,'
			.'near_edges,nodes,node_alignment,'
			.'orientation,outport_sharing,'
			.'pmin,pmax,portsharing,priorityphase,'
			.'randomfactor,randomimpulse,randomrounds,repulsion,rmin,rmax,'
			.'scaling,shape,shrink,smanhattan_edges,smax,splinefactor,splines,spreadlevel,state,stretch,straightphase,subgraph_labels,'
			.'tempfactor,tempmin,tempmax,tempscheme,temptreshold,textcolor,textmode,title,treefactor,'
			.'useractioncmd1,useractioncmd2,useractioncmd3,useractioncmd4,useractionname,'
			.'vertical_order,view,'
			.'width,'
			.'x,xbase,xmax,xlspace,xspace,'
			.'y,ybase,ymax,yspace,'
			.'class,range,'
			,",$cmd,")>0) return true;
			
		if ( ($state == 'IN_NODE' || $state =='IN_SUBGRAPH') && strpos(' ,'
			.'align,'
			.'bordercolor,borderstyle,borderwidth,'
			.'color,'
			.'focus,font,fontname,'
			.'height,hidden,horizontal_order,href,'
			.'icon,iconfile,importance,info1,info2,info3,invisible,'
			.'kind,'
			.'label,layer,level,loc,'
			.'partof,'
			.'scaling,shape,shrink,stretch,'
			.'textcolor,textmode,textwidth,title,type,'
			.'url,'
			.'vertical_order,'
			.'width,'
			,",$cmd,")>0) return true;
			
		if ( ($state == 'IN_EDGE') && strpos(' ,'
			.'align,anchor,arrowcolor,arrowsize,arrowstyle,arrow_mode,'
			.'backarrowstyle,backarrowsize,backarrowcolor,'
			.'classname,color,'
			.'font,fontname,'
			.'hidden,'
			.'invisible,'
			.'kind,'
			.'label,linestyle,'
			.'priority,'
			.'source,'
			.'target,thickness,textcolor,textmode,textwidth,type,'
			.'vertical_order,'
			,",$cmd,")>0) return true;
			
		if ( ($state == 'IN_NODETYPE') && strpos(' ,'
			.'align,'
			.'bordercolor,borderstyle,borderwidth,'
			.'color,'
			.'focus,font,fontname,'
			.'height,hidden,horizontal_order,href,'
			.'icon,iconfile,importance,info1,info2,info3,invisible,'
			.'label,layer,level,loc,'
			.'scaling,shape,shrink,'
			.'textcolor,textmode,textwidth,title,'
			.'url,'
			.'vertical_order,'
			.'width,weight,'
			,",$cmd,")>0) return true;
			
		if ( ($state == 'IN_EDGETYPE') && strpos(' ,'
			.'align,anchor,arrowcolor,arrowsize,arrowstyle,arrow_mode,'
			.'backarrowstyle,backarrowsize,backarrowcolor,'
			.'classname,color,'
			.'font,fontname,'
			.'hidden,'
			.'invisible,'
			.'kind,'
			.'label,linestyle,'
			.'priority,'
			.'thickness,textcolor,textmode,textwidth,title,'
			.'vertical_order,weight,'
			,",$cmd,")>0) return true;
			
		if ( ($state == 'IN_SUBGRAPH') && strpos(' ,'.
			'state,'
			,",$cmd,")>0) return true;

		if ( ($state == 'IN_REGION') && strpos(' ,'
			.'use_edgetypes,'
			.'range,'
			.'sources,state,'
			.'targets,'
			,",$cmd,")>0) return true;
			
		if ( ($state == 'BEGIN') && strpos(' ,'
			.'graph,%EOF,'
			,",$cmd,")>0) return true;
			
		return false;
	}
	
	// add a new command to the syntax tree
	public function add($cmd) {
		
		// $this->parser->errorText .= "cmd $cmd->name: '$cmd->arg'\n"; 			
		
		if (!$this->IsValidStateForCommand($cmd->name,$this->state)) {
			if ($cmd->name == '%EOF') {
				$this->error('unexpected end of file','');
			}
			else {
				$this->error('unexpected command; state = '.$this->state,$cmd);
			}
			return;
		}
		if ($cmd->name=='%EOF') {
			if ($this->state != 'BEGIN') {
				$cmd->name='open block';
				$this->error("unexpected end of script",$cmd);
				$this->state == 'BEGIN';
			}
			return;
		}
					
		if ($this->state == 'BEGIN') {
			if ($cmd->name == "graph") {
				$this->state = 'IN_GRAPH';
				$this->cmds[] = $cmd;
			}
			else {
				$this->error("'graph:' expected.",$cmd);
			}

		} else if ($this->state == 'IN_GRAPH') {
			if ($cmd->name=="subgraph") {
				$this->currentNode = array();
				$this->currentNode['kind'] = 'subgraph';
				$this->state= 'IN_SUBGRAPH';
			} else if ($cmd->name=="node") {
				$this->currentNode = array();
				$this->currentNode['kind'] = 'node';
				$this->state= 'IN_NODE';
				
			} else if (strpos(' ,'
				.'edge,backedge,nearedge,bentnearedge,leftnearedge,rightnearedge,leftbentnearedge,rightbentnearedge,'
				,",$cmd->name,") > 0) {
				$this->currentEdge = array();
				$this->state= 'IN_EDGE';
				
			} else if ($cmd->name=="nodetype") {
				$this->currentNodeType = array();
				$this->state= 'IN_NODETYPE';
				
			} else if ($cmd->name=="edgetype") {
				$this->currentEdgeType = array();
				$this->state= 'IN_EDGETYPE';
				
			} else if ($cmd->name=="region") {
				$this->currentRegion = array();
				$this->state= 'IN_REGION';
				
			} else {
				if ($cmd->name=='nodetypes') {
					$types = split(',',$cmd->arg);
					foreach ($types as $i => $type) {
						if ($type=='') continue;
						$this->nodeTypeSeq[$type] = 99999 - $i;
					}
				} else if ($cmd->name=='edgetypes') {
					$types = split(',',$cmd->arg);
					foreach ($types as $i => $type) {
						if ($type=='') continue;
						$this->edgeTypeSeq[$type] = 99999 - $i;
					}
				} else if ($cmd->name=='show_nodetypes') {
					$this->filters[0] .= $cmd->arg;
				} else if ($cmd->name=='show_edgetypes') {
					$this->filters[1] .= $cmd->arg;
				} else if ($cmd->name=='skip_nodetypes') {
					$this->filters[2] .= $cmd->arg;
				} else if ($cmd->name=='skip_edgetypes') {
					$this->filters[3] .= $cmd->arg;
				} else if ($cmd->name=='hide_nodetypes') {
					$this->filters[4] .= $cmd->arg;
				} else if ($cmd->name=='hide_edgetypes') {
					$this->filters[5] .= $cmd->arg;
				} else if ($cmd->name=='show_nodes') {
					$this->filters[6] .= $cmd->arg;
				} else if ($cmd->name=='skip_nodes') {
					$this->filters[7] .= $cmd->arg;
				} else if ($cmd->name=='hide_nodes') {
					$this->filters[8] .= $cmd->arg;
				}
				else if ($cmd->name == '}') {
					// we change the default for edge labels
					if (!$this->edgeLabelsCmd) $this->cmds[] = new Command("display_edge_labels","","boolean","yes","");
					$this->cmds[] = $cmd;
					$this->state='BEGIN';
				} else {
					if ($cmd->name=='display_edge_labels') $this->edgeLabelsCmd = true;
					$this->cmds[] = $cmd;
				}
			}

			
		} else if ($this->state == 'IN_NODE' || $this->state == 'IN_SUBGRAPH') {
			if ($cmd->name == '}') {
				if (!array_key_exists('title',$this->currentNode)) {
					$this->error("title missing for node / subgraph ",$cmd);
				}
				else {
					$title = $this->currentNode['title'];
					if (array_key_exists($title,$this->nodes)) {
						$this->nodes[$title] = array_merge($this->nodes[$title],$this->currentNode);
					}
					else {
						$this->nodes[$title] = $this->currentNode;
					}
					if ($this->state == 'IN_SUBGRAPH') $this->partOf[$title] = $this->currentNode['partof'];
				}
				$this->state = 'IN_GRAPH';
			}
			else {
				// if the node label contains "%%" we try to substitute this symbol by a previously defined label
				if ($cmd->name=='label' && array_key_exists('title',$this->currentNode) &&
					array_key_exists($this->currentNode['title'],$this->nodes) &&
					array_key_exists('label',$this->nodes[$this->currentNode['title']]) ) {
						$this->nodes[$this->currentNode['title']]['label']= 
							str_replace('%%',$this->nodes[$this->currentNode['title']]['label'],$cmd->arg);
				} else if ($cmd->name=='type' && array_key_exists('title',$this->currentNode) &&
					array_key_exists($this->currentNode['title'],$this->nodes) &&
					array_key_exists('type',$this->nodes[$this->currentNode['title']]) ) {
						$this->nodes[$this->currentNode['title']]['type'].=$cmd->arg;
						// $this->parser->errorText = "type='".$this->nodes[$this->currentNode['title']]['type']."'\n";
				} else {
					// assign the property
					$this->currentNode[$cmd->name] = $cmd->arg;
				}					
									
				// clear label and bordercolor if iconfile is used
				if ($cmd->name =='iconfile' || $cmd->name == 'icon') {
					if (!array_key_exists('label',$this->currentNode)) $this->currentNode['label']='';
					if (!array_key_exists('bordercolor',$this->currentNode)) $this->currentNode['bordercolor']='white';
				}
			}
			
		} else if ($this->state == 'IN_EDGE') {
			if ($cmd->name == '}') {
				if (!array_key_exists('source',$this->currentEdge)
				  ||!array_key_exists('target',$this->currentEdge)) {
					$this->error("source/from and target/to are needed for an edge definition",$cmd);
				}
				else {
					// register the edge and make some calculations
					$this->edges[++$this->edgeNumber] = $this->currentEdge;
					$this->edgeHasNode[$this->edgeNumber][$this->currentEdge['source']] = true;
					$this->edgeHasNode[$this->edgeNumber][$this->currentEdge['target']] = true;
					$this->countEdgesPerNode($this->currentEdge,$this->currentEdge['source']);
					$this->countEdgesPerNode($this->currentEdge,$this->currentEdge['target']);
				}
				$this->state = 'IN_GRAPH';
			}
			else {
				if ($cmd->name=='hidden' || ($cmd->name=='invisible' && $cmd->arg=="yes")) {
					$this->currentEdge['linestyle'] = 'invisible';
					$this->currentEdge['textcolor'] = 'white';
				}
				else {
					// assign the property
					$this->currentEdge[$cmd->name] = $cmd->arg;
				}
			}
			
		} else if ($this->state == 'IN_NODETYPE') {
			if ($cmd->name == '}') {
				if (!array_key_exists('title',$this->currentNodeType)) {
					$this->error("title missing for nodetype",$cmd);
				}
				else {
					$title = $this->currentNodeType['title'];
					$this->nodeTypes[$title] = $this->currentNodeType;
					if (!array_key_exists($title,$this->nodeTypeSeq)) {
						if (array_key_exists('weight',$this->currentNodeType)) {
							$this->nodeTypeSeq[$title] = $this->currentNodeType['weight'];
						}
						else {
							$this->nodeTypeSeq[$title] = 100- (++$this->nodeTypeNumber);
						}
					}
				}
				$this->state = 'IN_GRAPH';
			}
			else {
				// assign the property
				$this->currentNodeType[$cmd->name] = $cmd->arg;
				// clear label and bordercolor if iconfile is used
				if ($cmd->name =='iconfile' || $cmd->name=='icon') {
					if (!array_key_exists('label',$this->currentNodeType)) $this->currentNodeType['label']='';
					if (!array_key_exists('bordercolor',$this->currentNodeType)) $this->currentNodeType['bordercolor']='white';
				}
			}
			
		} else if ($this->state == 'IN_EDGETYPE') {
			if ($cmd->name == '}') {
				if (!array_key_exists('title',$this->currentEdgeType)) {
					$this->error("title missing for edgetype",$cmd);
				}
				else {
					$this->currentEdgeType['class']= ++$this->edgeClassNumber;
					$title = $this->currentEdgeType['title'];
					$this->edgeTypes[$title] = $this->currentEdgeType;
					if (!array_key_exists($title,$this->edgeTypeSeq)) {
						if (array_key_exists('weight',$this->currentEdgeType)) {
							$this->edgeTypeSeq[$title] = $this->currentEdgeType['weight'];
						}
						else {
							$this->edgeTypeSeq[$title] = 100 - (++$this->edgeTypeNumber);
						}
					}
				}
				$this->state = 'IN_GRAPH';
				$this->add(new Command('classname',$this->edgeClassNumber,'string',$title,$cmd->pos));
			}
			else {
				$this->currentEdgeType[$cmd->name] = $cmd->arg;
			}
			
		} else if ($this->state == 'IN_REGION') {
			if ($cmd->name == '}') {
				if (!array_key_exists('state',$this->currentRegion)) $this->currentRegion['state']='boxed';
				if      (array_key_exists('range',$this->currentRegion) && array_key_exists('targets',$this->currentRegion)) {
					$this->error("a region cannot have 'range' and 'targets' at the same time",$cmd);
				}
				else if (!array_key_exists('sources',$this->currentRegion) || preg_match('/^[, ]*$/',$this->currentRegion['sources'])) {
					$this->error("a region must have 'sources'",$cmd);
				}
				else if (strpos(' ,boxed,folded,exclusive,',$this->currentRegion['state']) <= 0) {
					$this->error("region state can only be 'boxed', 'folded' and 'exclusive'",$cmd);
				}
				else {
					// register the region
					$this->regions[++$this->regionNumber] = $this->currentRegion;
				}
				$this->state = 'IN_GRAPH';
			}						
			else {
				$this->currentRegion[$cmd->name] = $cmd->arg;
			}
			
		}
		
	}

	public function countEdgesPerNode($edge,$nodeName) {
		if (!array_key_exists('invisible',$edge) || $edge['invisible']!='yes') {
			if (!array_key_exists($nodeName,$this->nodeEdgeCount)) {
				$this->nodeEdgeCount[$nodeName] = 1;
			} 
			else ++ $this->nodeEdgeCount[$nodeName];
		}
	}
	
	public function createUndefinedNode($graphName,$nodeName) {
		if ($nodeName=='' || array_key_exists($nodeName,$this->nodes)) return;
		$node = array();
		$node['title'] = $nodeName;
		$node['label'] = "$nodeName";
		$node['partof'] = $graphName;
		$node['type'] = ',undefined node,';
		$node['kind'] = 'normal';
		$this->nodes[$nodeName]= $node;
	}
		
	public function findUndefinedNodes($graphName) {
		foreach ($this->edges as $edge) {
			$nodeName = $edge['source'];
			$this->createUndefinedNode($graphName,$nodeName);
			$nodeName = $edge['target'];
			$this->createUndefinedNode($graphName,$nodeName);
		}
		foreach ($this->regions as $region) {
			if (array_key_exists('sources',$region)) {
				$objs = split(',',$region['sources']);
				foreach ($objs as $obj) $this->createUndefinedNode($graphName,$obj);
			}
			if (array_key_exists('targets',$region)) {
				$objs = split(',',$region['targets']);
				foreach ($objs as $obj) $this->createUndefinedNode($graphName,$obj);
			}
		}
	}
	
		
	//====================================================================
	// the "do it all" function -- should be split into parts
	public function resolveTypes($graphName) {

		// projection of nodetype related attributes into nodes
		array_multisort($this->nodeTypeSeq,SORT_NUMERIC,SORT_DESC);
		foreach ($this->nodeTypeSeq as $nodeType => $weight) {
			// $this->parser->errorText .= "type: " . $nodeType . "\n";
			foreach ($this->nodes as $nodeTitle => $node) {
				if (!array_key_exists('type',$node)) continue;
				if (strpos($node['type'],','.$nodeType.',') === false) continue;
				foreach ($this->nodeTypes[$nodeType] as $atrib => $value) {
					if ($atrib=="weight" || $atrib=="title") continue;
					if (array_key_exists($atrib,$node)) {
						// we do not overwrite existing attributes except the label
						if ($atrib=='label') {
							// if the nodetype label contains "%%" we substitute this symbol by the current label
							$this->nodes[$nodeTitle]['label']= str_replace('%%',$this->nodes[$nodeTitle]['label'],$value);
						}
					}
					else {
						$this->nodes[$nodeTitle][$atrib] = $value;
						// $this->parser->errorText .= "node: $nodeTitle , types = ".$node['type']." -- ".$atrib ."=".$value."\n";
					}
				}
			}
		}		
		
		// projection of edgetype related attributes into edges
		array_multisort($this->edgeTypeSeq,SORT_NUMERIC,SORT_DESC);
		foreach ($this->edgeTypeSeq as $edgeType => $weight) {
			//$this->parser->errorText .= "type: " . $edgeType . "\n";
			foreach ($this->edges as $edgeId => $edge) {
				if (!array_key_exists('type',$edge)) continue;
				if (strpos($edge['type'],','.$edgeType.',') === false) continue;
				foreach ($this->edgeTypes[$edgeType] as $atrib => $value) {
					if ($atrib=="weight" || $atrib=="title") continue;
					if ($atrib=='kind') {
						if ($edge['kind']!='normal') continue;
					}
					else if (array_key_exists($atrib,$edge)) continue;
					// we do not overwrite existing attributes except the label
					// $this->parser->errorText .= "edge: $edgeTitle , types = ".$edge['type']." -- ".$atrib ."=".$value."\n";
					if ($atrib=='label') {
						// if the edgetype label contains "%%" we substitute this symbol by the current label
						$this->edges[$edgeId]['label']= str_replace('%%',$this->edges[$edgeId]['label'],$value);
					} else {
						$this->edges[$edgeId][$atrib] = $value;
					}
				}
			}
		}
		
		// prepare filter arrays for nodes and edges
		for($f=0;$f<9;$f++) {
			$condition = split(',',$this->filters[$f]);
			$this->filters[$f] = array();
			foreach ($condition as $cond) {
				if ($cond!='') $this->filters[$f][$cond] = true;
			}
		}
		
		// check visibility of edges and make nodes invisible that are not connected to a visible edge
		
		foreach ($this->edges as $edgeId => $edge) {
			if ( $this->hasType($this->filters[5],$edge)) {
				if (--$this->nodeEdgeCount[$edge['source']] <=0 ) $this->nodes[$edge['source']]['invisible'] = 'yes';
				if (--$this->nodeEdgeCount[$edge['target']] <=0 ) $this->nodes[$edge['target']]['invisible'] = 'yes';
			}
			if ( $this->hasType($this->filters[3],$edge ) ||
			    (count($this->filters[1]) > 0 && !$this->hasType($this->filters[1],$edge))) {
				if (--$this->nodeEdgeCount[$edge['source']] <=0 ) $this->nodes[$edge['source']] = array();
				if (--$this->nodeEdgeCount[$edge['target']] <=0 ) $this->nodes[$edge['target']] = array();
			}
		}
		
		
		$theEnd = array_pop($this->cmds);
		
		// output NODES 
		// we flip the list, thus (assuming orientation top_to_bottom) aiSee will place 
		// the nodes from left to right in the order they were defined
		
		$this->nodes = array_reverse($this->nodes);
		
		$this->outputNodesForGraph($graphName);		

		// filter EDGES

		$rEdges = array(); // the edges remaining after applying filters
		$nEdges = array(); // the near edges remaining after applying filters
				
		foreach ($this->edges as $edgeId => $edge) {
			// ignore edges which were removed because of skipped nodes
			if (count($edge) == 0) continue;
			
			// check edge against filter			
			if ( $this->hasType($this->filters[5],$edge)) {
				$edge['linestyle'] = 'invisible';
				$edge['textcolor'] = 'white';
			}
			if ( $this->hasType($this->filters[3],$edge ) ||
			    (count($this->filters[1]) > 0 && !$this->hasType($this->filters[1],$edge))) {
				continue;
			}

			$rEdges[$edgeId] = $edge;
			// collect bear edges for special check
			if ($edge['kind'] != 'normal') $nEdges[$edgeId] = $edge;
		}

		// check conflicts with multiple near edges
		$this->checkNearEdgeConflicts($nEdges);
			

		// output EDGES

		foreach ($rEdges as $edgeId => $edge) {
			
			$from = $edge['source'];
			$to   = $edge['target'];
			$edgeKind= $edge['kind'];
			// $this->parser->errorText .= $from . "--".$to."--".$edgeKind."\n";			
			if (array_key_exists($from,$this->nodes) && 
				array_key_exists('layer',$this->nodes[$from])) 	$fromLayer = $this->nodes[$from]['layer'];
			else 										    	$fromLayer ='?';
			if (array_key_exists($to,$this->nodes) &&
				array_key_exists('layer',$this->nodes[$to]))   	$toLayer = $this->nodes[$to]['layer'];
			else 											    $toLayer ='?';
			if 		($edgeKind=='back') 		$this->cmds[]=new Command("backedge","","","{","");
			else if ($edgeKind=='near') 		$this->cmds[]=new Command("nearedge","","","{","");
			else if ($edgeKind=='bent') 		$this->cmds[]=new Command("bentnearedge","","","{","");
			else if ($edgeKind=='rightbentnear')$this->cmds[]=new Command("rightbentnearedge","","","{","");
			else if ($edgeKind=='leftbentnear')	$this->cmds[]=new Command("leftbentnearedge","","","{","");
			else if ($edgeKind=='rightnear')	$this->cmds[]=new Command("rightnearedge","","","{","");
			else if ($edgeKind=='leftnear')		$this->cmds[]=new Command("leftnearedge","","","{","");
			else {
				if ($fromLayer <= $toLayer) {
					$this->cmds[]=new Command("edge","","","{","");
				}
				else {
					// normal edge, swap to backedge to accomodate to layers
					$this->edges[$edgeId]['source'] = $to;
					$this->edges[$edgeId]['target'] = $from;
					$this->cmds[]=new Command("backedge","","","{","");				
				}
			}
			if (array_key_exists('label',$edge)) {
				if (array_key_exists('textwidth',$edge)) $edge['label'] = $this->breakLabel($edge['label'],$edge['textwidth']);
				else									 $edge['label'] = $this->breakLabel($edge['label']);
			}
			foreach ($edge as $atrib => $value) {
				if ($atrib=='back') continue;
				$this->cmds[]=new Command($atrib,"",$this->parser->hasArgumentType($atrib),$value,"");
			}
			$this->cmds[]=new Command("}","","","","");
		}

			
		// output REGIONS
		
		foreach ($this->regions as $region) {
			$this->cmds[]=new Command("region","","","{","");
			foreach ($region as $atrib => $value) {
				$type = $this->parser->hasArgumentType($atrib);
				if ($atrib =='sources') $atrib= 'source';
				if ($atrib =='targets') $atrib= 'target';
				if ($atrib =='use_edgetypes') {
					$atrib= 'class';
					$etypes = split(',',$value);
					$value='';
					foreach ($etypes as $etype) {
						if ($etype=='') continue;
						if (!array_key_exists($etype,$this->edgeTypes)) {
							$this->error("unknown edgetype '$etype' in region",'');
						} else {
							$value .= $this->edgeTypes[$etype]['class']." ";
						}
			}
					
				}
				$this->cmds[]=new Command($atrib,"",$type,$value,"");
			}
			$this->cmds[]=new Command("}","","","","");
		}
		
		
		$this->cmds[] = $theEnd;
		
		// mark commands as obsolete which were only needed on WGL level
		foreach ($this->cmds as $cmd) {
			if (	$cmd->name=="weight"
				||	$cmd->name=="layer"
				||	$cmd->name=="type"
				||	$cmd->name=="url"
				||	$cmd->name=="href"
				||	$cmd->name=="textwidth"
				||	$cmd->name=="kind"
				||	$cmd->name=="partof"
				) {
				$cmd->name = "#".$cmd->name;
			}
		}
	}

	// output nodes which belong to a given (sub)graph
	public function outputNodesForGraph($graph) {
		// $this->parser->errorText .= "enter graph $graph\n";
		foreach ($this->nodes as $nodeName => $node) {
			// ignore nodes which were removed because of skipped edges
			if (count($node) == 0) continue;

			// only nodes of this graph
			if ($node['partof'] != $graph) continue;
			
			// check node against filter			
			if ( $this->hasType($this->filters[4],$node) || array_key_exists($nodeName,$this->filters[8]) ) {
				$node['invisible'] = 'yes';
				$this->hideEdgesForNode($node['title']);
			}
			if ( $this->hasType($this->filters[2],$node ) || array_key_exists($nodeName,$this->filters[7])
				 || (count($this->filters[0]) > 0 && !$this->hasType($this->filters[0],$node))
			     || (count($this->filters[6]) > 0 && !array_key_exists($nodeName,$this->filters[6])) ) {
				$this->removeEdgesForNode($node['title']);
				continue;
			}
			//$this->parser->errorText .= "checking graph $graph : node $nodeName \n";

			if ($node['kind']=='subgraph')	$this->cmds[]=new Command("graph","","","{","");
			else 							$this->cmds[]=new Command("node","","","{","");
			
			// output node properties, also for subgraph nodes
			if (array_key_exists('url',$node) && ! array_key_exists('info3',$node)) {
				$this->cmds[]=new Command("info3","","string",
				"href:../../index.php?title=" . urlencode($node['url']).";" ,"");
			}
			if (array_key_exists('href',$node) && ! array_key_exists('info3',$node)) {
				$this->cmds[]=new Command("info3","","string",
				"href:../../index.php?title=" . urlencode($node['href']).";" ,"");
			}
			if (array_key_exists('label',$node)) {
				if (array_key_exists('textwidth',$node)) $node['label'] = $this->breakLabel($node['label'],$node['textwidth']);
				else									 $node['label'] = $this->breakLabel($node['label']);
			}			  
			foreach ($node as $atrib => $value) {
				$this->cmds[]=new Command($atrib,"",$this->parser->hasArgumentType($atrib),$value,"");
			}
						
			if ($node['kind']=='subgraph') {
				// output nodes of nested subgraph
				$this->outputNodesForGraph($nodeName);
			}
						
			// close block			
			$this->cmds[]=new Command("}","","","","");
			
		}
	}
	
	
	
	// apply some simple word wrapping to texts and replace wiki '' by GDL notation
	public function breakLabel($label, $textWidth=15) {
		$n = 0;
		$underline = false;
		$italic	= false;
		$bold	= false;
		$label=preg_replace(";</[^>]*>;",'',$label);
		$label=preg_replace(";<br[^>]*>;","\n",$label);
		$label=preg_replace(";<p[^>]*>;","\n",$label);
		$tt='';
		for ($i=0; $i<strlen($label); $i++) {
			$c = $label[$i];
			if ($c=="<") {
				if (!$underline && strpos($label,"<u>",$i)==$i) {
					$underline=true;
					$tt.="\\fu";
					$i+=2;
				} else if ($underline && strpos($label,"</u>",$i)==$i) {
					$underline=false;
					$tt.="\\fn";
					if ($italic) $tt .="\\fI";
					if ($bold)   $tt .="\\fb";
					$i+=3;
				} else $tt.=$c;
				continue;
			}
			if ($n==0) {
				if ($c=="'") $n=1;
				else 		 $tt.=$c;
			} else if ($n==1) {
				if ($c=="'") {
					$n=2;
				}
				else {
					$tt.="'$c";
					$n=0;
				}
			} else if ($n==2) {
				if ($c=="'") {
					$n=3;
				}
				else {
					if ($italic) {
						$tt.="\\fn$c";
						$italic = false;
						if ($bold) $tt .="\\fb";
						if ($underline) $tt .="\\fu";
					}
					else {
						$tt.="\\fI$c";
						$italic = true;
					}
					$n=0;
				}
			} else if ($n==3) {
				if ($bold) {
					$tt.="\\fn";
					$bold = false;
					if ($italic) $tt .="\\fI";
					if ($underline) $tt .="\\fu";
				}
				else {
					$tt.="\\fb";
					$bold = true;
				}
				$n=0;
				if ($c=="'") $n=1;
				else		 $tt.=$c;
			}
		}
		if ($n>0) $tt.="\\fn";
		// $tt = str_replace('----',"\\fi045");  // horizontal line within node

		$tt = str_replace("\\n","\n",$tt);
		$t ='';

		for ($i=0; $i<strlen($tt); $i++) {
			$c = $tt[$i];
			if ($n==0 && $c==' ') continue;
			if ($c=="\n") {
				$t.=$c;
				$n=0;
			} else if ($n>=$textWidth) {
				if ($c==' ') {
					$t.="\n";
					$n= 0;
				} else if ($c==':' || $c=='.' || $c==',' || $c==';' || $c=='-' || $c=='/') {
					$t.= "$c\n";
					$n= 0;
				} else if ($c=='(' || $c=='[' || $c=='{') {
					$t.= "\n$c";
					$n= 1;
				}
				else {
					$t.= $c;
					$n++;
				}
			} else {
				$t.= $c;
				$n++;
			}
		}
		return $t;
	}

	// check special restrictions on near edges
	public function checkNearEdgeConflicts($edges) {
		foreach ($edges as $edgeId => $edge) {
			$n=$ln=$rn=$b=$lb=$rb=0;
			$edgeKind = $edge['kind'];
			if 		($edgeKind == 'near') 			$n++;
			else if ($edgeKind == 'leftnear') 		$ln++;
			else if ($edgeKind == 'rightnear') 		$rn++;
			else if ($edgeKind == 'bentnear')	 	$b++;
			else if ($edgeKind == 'leftbentnear') 	$lb++;
			else if ($edgeKind == 'rightbentnear') 	$rb++;
			
			foreach ($edges as $nEdgeId => $nEdge) {
				$from = $edge['source'];
				if ($nEdgeId <= $edgeId || $nEdge['source'] != $from) continue;
				if ($nEdge['target'] == $edge['target']) {
					$this->error("near edges at '$from' point to the same other node '".$edge['target'],'');
				}
				$nEdgeKind = $nEdge['kind'];
				if 		($nEdgeKind == 'near') 			$n++;
				else if ($nEdgeKind == 'leftnear') 		$ln++;
				else if ($nEdgeKind == 'rightnear') 	$rn++;
				else if ($nEdgeKind == 'bentnear')	 	$b++;
				else if ($nEdgeKind == 'leftbentnear') 	$lb++;
				else if ($nEdgeKind == 'rightbentnear')	$rb++;

				//$this->parser->errorText .= "checking $edgeKind edge $edgeId against $nEdgeKind edge $nEdgeId at $from: $n,$ln,$rn  $b,$lb,$rb:\n";			
				
				//if (($n+$ln+$rn+$b+$lb+$rb)>2) $this->error("too many near edges at '$from'; maximum = 2",'');
				//if (($ln>=1 && $lb>=1) || $ln>=2 || $lb>=2 ) $this->error("too many left near edges at '$from'; maximum = 1",'');
				//if (($rn>=1 && $rb>=1) || $rn>=2 || $rb>=2 ) $this->error("too many right near edges at '$from'; maximum = 1",'');
				if (($rn+$ln>=2 && $rb+$lb>=1) ) $this->error("too many near (bent) edges at '$from'",'');
			}
		}
	}
	
	// learn color mappings, convert into colorentry commands
	public function useColorMap($colorMap) {
		$this->colorEntries = array();
		$hexMap['a'] = 10; $hexMap['b'] = 11; $hexMap['c'] = 12;
		$hexMap['d'] = 13; $hexMap['e'] = 14; $hexMap['f'] = 15;
		for ($i=0;$i<10;$i++) $hexMap[$i] = $i;
		foreach ($colorMap as $col => $colNr) {
			$r = 16 * $hexMap[$col[1]] + $hexMap[$col[2]];
			$g = 16 * $hexMap[$col[3]] + $hexMap[$col[4]];
			$b = 16 * $hexMap[$col[5]] + $hexMap[$col[6]];
			$colTriple = "$r $g $b";
			$this->colorEntries[] = new Command("colorentry",$colNr,"colorset",$colTriple,"");
		}		
	}
	
	
	// output the syntax tree
	public function printTreeAsGdl() {
		$text="";
		global $wgLang;
		foreach ($this->cmds as $cmdNr => $cmd) {
			if ($cmdNr==1) {   // insert colorentries at top of graph definition
				$text .= "// generated by Wgraph at "
						.$wgLang->timeanddate(wfTimeStampNow(),true)
						."\n// see http://semeb.com/dpldemo/index.php/Wgraph\n";
				foreach ($this->colorEntries as $colorEntry) {
					$text .= $colorEntry->show();
				}
			}
			$text .= $cmd->show();
		}
		return $text;
	}
		
	// print an error message
	public function error($msg, $cmd) {
		if (!is_object($cmd))	$this->parser->errorText.= "error : $msg\n";
		else					$this->parser->errorText.= "error ($cmd->pos): $msg\n--   '$cmd->name'\n";
	}
		
	// check if node or edge belongs to one of a given set of types
	public function hasType($types,$object) {
		if (count($types)<=0 || !array_key_exists('type',$object))	return false;
		foreach ($types as $type => $dummy) {
			if (strpos($object['type'],",$type,") === false) continue;
			return true;
		}
		return false;
	}
		
	// destroy all edges which touch a given node
	public function removeEdgesForNode($nodeName) {
		foreach ($this->edges as $edgeNr => $edge) {
			if (array_key_exists($nodeName,$this->edgeHasNode[$edgeNr])) {
				$this->edges[$edgeNr] = array();
			}
		}
	}

	// make all edges invisible which touch a given node
	public function hideEdgesForNode($nodeName) {
		foreach ($this->edges as $edgeNr => $edge) {
			if (array_key_exists($nodeName,$this->edgeHasNode[$edgeNr])) {
				$this->edges[$edgeNr]['linestyle'] = 'invisible';
				$this->edges[$edgeNr]['textcolor'] = 'white';
			}
		}
	}
}

//*******************************************************************************************************//

class Command {
	public $name;
	public $inx;
	public $type;
	public $arg;
	public $pos;
	
	function Command($name, $inx, $type, $arg, $pos) {
		// allow 'font' as an alias for 'fontname'
		if      ($name=='font') $name='fontname';
		else if ($name=='align') {
			$name='textmode';
			if      ($arg=='left')  $arg='left_justify';
			else if ($arg=='right') $arg='right_justify';
		}
		else if ($name=='icon' || $name=='iconfile') {
			// add default extension ".png"
			$name='iconfile';
			if (strpos($arg,'.')===false)  $arg.='.png';
		}
		
		$this->name = $name;
		$this->inx  = $inx;
		$this->type = $type;
		$this->arg  = $arg;
		$this->pos  = $pos;
	}

	function show() {
		if ($this->name[0]=="#") return ""; // return "/* $this->name $this->index $this->arg */\n";
		if ($this->name == "}") return "}\n";
		if ($this->type=="string") {
			$arg = '"'.$this->arg.'"';
		} else if ($this->type=='objlist') {
			$objs = split(',',$this->arg);
			$arg='';
			foreach ($objs as $obj) {
				if ($obj=='') continue;
				$arg .= "\"$obj\" ";
			}
		} else if ($this->type=='numberlist') {
			$objs = split(',',$this->arg);
			$arg='';
			foreach ($objs as $obj) {
				if ($obj=='') continue;
				$arg .= "$obj ";
			}
		} else						$arg = $this->arg;
		if ($arg == "") return $this->name;
		if ($this->inx != '') return "$this->name $this->inx: $arg\n";
		return "$this->name: $arg\n";
	}
	
}

//*******************************************************************************************************//

?>